import sys
from typing import Optional


def read_command_line_arg(argument: str) -> Optional[str]:
    """Read an command line argument.

            Arguments:
                argument -- name of the cli argument to get the value of, will auto append "--" to it
            Returns:
                the value of the argument, None if doesn't exist
    """
    if command_line_arg_defined(argument) is True:
        arg_list = sys.argv
        key_index = arg_list.index("--" + argument) + 1
        reply = arg_list[key_index]
    else:
        reply = None
    return reply


def command_line_arg_defined(argument: str) -> bool:
    """Check if a command line argument is defined.

            Arguments:
                argument -- name of the cli argument to get the value of, will auto append "--" to it
            Returns:
                True if argument is declared, False otherwise
    """

    if "--" + argument in sys.argv:
        return True
    else:
        return False


def read_all_cli_args_to_dict() -> dict:
    """Returns all cli args (that start with --) key/value pairs as a single dict.

            Returns:
                argument_dict -- A dict of all cli arguments key/value pair
        """
    argument_dict = {}
    arg_list = sys.argv
    for argument in arg_list:
        if argument.startswith("--"):
            key_index = arg_list.index(argument) + 1
            argument_dict[argument[2:]] = arg_list[key_index]
    return argument_dict



from unittest import TestCase, mock
from parse_it import ParseIt
from parse_it.command_line_args.command_line_args import *
from parse_it.envvars.envvars import *
from parse_it.file.toml import *
from parse_it.file.hcl import *
from parse_it.file.yaml import *
from parse_it.file.json import *
from parse_it.file.env import *
from parse_it.file.ini import *
from parse_it.file.xml import *
from parse_it.file.file_reader import *
from parse_it.type_estimate.type_estimate import *
import os


VALID_FILE_TYPE_EXTENSIONS = [
    "json",
    "yaml",
    "yml",
    "toml",
    "tml",
    "conf",
    "cfg",
    "ini",
    "xml"
]

test_files_location = os.getenv("TEST_FILES_LOCATION", "test_files")


class BaseTests(TestCase):

    maxDiff = None

    def test_command_line_args_empty(self):
        reply = read_command_line_arg("empty_variable")
        self.assertIsNone(reply)

    def test_command_line_arg_defined_false(self):
        reply = command_line_arg_defined("empty_variable")
        self.assertFalse(reply)

    def test_envvars_empty(self):
        reply = read_envvar("empty_variable")
        self.assertIsNone(reply)

    def test_envvars_string(self):
        os.environ["TEST_ENVVAR"] = "TEST_ENVVAR_VALUE"
        reply = read_envvar("TEST_ENVVAR")
        self.assertEqual(reply, "TEST_ENVVAR_VALUE")

    def test_envvars_string_forced_uppercase(self):
        os.environ["TEST_ENVVAR_UPPERCASE"] = "TEST_ENVVAR_UPPERCASE_VALUE"
        reply = read_envvar("test_envvar_uppercase", force_uppercase=True)
        self.assertEqual(reply, "TEST_ENVVAR_UPPERCASE_VALUE")

    def test_envvars_string_forced_lowercase(self):
        os.environ["test_envvvar_lowercase"] = "test_envvvar_lowercase_value"
        reply = read_envvar("test_envvvar_lowercase", force_uppercase=False)
        self.assertEqual(reply, "test_envvvar_lowercase_value")

    def test_file_reader_read_file(self):
        reply = read_file(test_files_location + "/test_read_file")
        self.assertEqual(reply, "it_reads!")

    def test_file_reader_read_file_does_not_exist(self):
        reply = read_file(test_files_location + "/totally_bogus_file")
        self.assertIsNone(reply)

    def test_file_reader_file_or_folder_read_folder(self):
        reply = file_or_folder(test_files_location)
        self.assertEqual(reply, "folder")

    def test_file_reader_file_or_folder_read_file(self):
        reply = file_or_folder(test_files_location + "/test.json")
        self.assertEqual(reply, "file")

    def test_file_reader_file_or_folder_read_no_file(self):
        reply = file_or_folder("/non_existing_path")
        self.assertIsNone(reply)

    def test_file_reader_folder_exists(self):
        reply = folder_exists(test_files_location)
        self.assertTrue(reply)

    def test_file_reader_folder_does_not_exists(self):
        reply = folder_exists("totally_bogus_directory")
        self.assertFalse(reply)

    def test_file_reader_strip_trailing_slash(self):
        reply = strip_trailing_slash(test_files_location + "/")
        self.assertEqual(reply, test_files_location)

    def test_file_reader_strip_trailing_slash_root_folder(self):
        reply = strip_trailing_slash("/")
        self.assertEqual(reply, "/")

    def test_file_reader_strip_trailing_slash_no_strip_needed(self):
        reply = strip_trailing_slash(test_files_location)
        self.assertEqual(reply, test_files_location)

    def test_file_reader_file_types_in_folder_folder_does_not_exist_raise_warn(self):
        with self.assertWarns(Warning):
            file_types_in_folder("totally_bogus_folder_location", VALID_FILE_TYPE_EXTENSIONS)

    def test_read_envvar_folder_does_not_exist_raise_warn(self):
        with self.assertWarns(Warning):
            os.environ["TEST_CONFIG_FOLDER_NON_EXISTING_ENVVAR"] = "TEST_CONFIG_FOLDER_NON_EXISTING_ENVVAR"
            parser = ParseIt(config_location="totally_bogus_folder_location", config_type_priority=[
                "yaml",
                "json",
                "env_vars"
            ])
            reply = parser.read_configuration_variable("test_config_folder_non_existing_envvar")
            self.assertEqual(reply, "TEST_CONFIG_FOLDER_NON_EXISTING_ENVVAR")

    def test_read_envvar_single_file_config(self):
        os.environ["FILE_TYPE"] = "envvar"
        parser = ParseIt(config_location=test_files_location + "/test.hcl")
        reply = parser.read_configuration_variable("file_type")
        self.assertEqual(reply, "envvar")

    def test_read_cli_args_folder_does_not_exist_raise_warn(self):
        with self.assertWarns(Warning):
            ParseIt(config_location="totally_bogus_folder_location", config_type_priority=[
                "yaml",
                "json",
                "cli_args"
            ])
            test_args = ["parse_it_mock_script.py", "--test_cli_key_no_folder", "test_value"]
            with mock.patch('sys.argv', test_args):
                parser = ParseIt()
                reply = parser.read_configuration_variable("test_cli_key_no_folder")
                self.assertEqual(reply, "test_value")

    def test_file_reader_file_types_in_folder(self):
        reply = file_types_in_folder(test_files_location, VALID_FILE_TYPE_EXTENSIONS)
        expected_reply = {
            'json': [
                'test.json',
                'test_none_values.json',
                'test_subfolder_1/test_sub_subfolder_2/test_subfolder_2.json',
                'test_subfolder_1/test_sub_subfolder_3/test_subfolder_3.json',
                'test_subfolder_1/test_sub_subfolder_3/test_subfolder_4.json',
                'test_subfolder_1/test_subfolder_1.json'
            ],
            'yaml': [
                'test.yaml'
            ],
            'yml': [],
            'toml': [
                'test.toml'
            ],
            'tml': [],
            'conf': [],
            'cfg': [],
            'ini': [
                'test.ini'
            ],
            'xml': [
                'test.xml'
            ]
        }
        self.assertDictEqual(reply, expected_reply)

    def test_ini(self):
        reply = parse_ini_file(test_files_location + "/test.ini")
        expected_reply = {
            'DEFAULT': {
                'file_type': "ini",
                'test_string': 'testing',
                'test_bool_true': 'true',
                'test_bool_false': 'false',
                'test_int': '123.0',
                'test_float': '123.123',
                'test_list': '["test1", "test2", "test3"]'
            },
            'test_ini': {
                'test_ini_key': 'test_ini_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_json(self):
        reply = parse_json_file(test_files_location + "/test.json")
        expected_reply = {
            'file_type': "json",
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_json': {
                'test_json_key': 'test_json_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_env_file(self):
        reply = parse_env_file(test_files_location + "/test.env")
        expected_reply = {
            'file_type': 'env',
            'test_string': 'testing',
            'test_bool_true': 'true',
            'test_bool_false': 'false',
            'test_int': '123',
            'test_float': '123.123',
            'test_list': '["test1","test2","test3"]',
            'test_env': '{"test_env_key": "test_env_value"}'}
        self.assertDictEqual(reply, expected_reply)

    def test_hcl(self):
        reply = parse_hcl_file(test_files_location + "/test.hcl")
        expected_reply = {
            'file_type': "hcl",
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_dict': {
                "hcl_dict_key": "hcl_dict_value"
            },
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_hcl': {
                'test_hcl_name': {
                    'test_hcl_key': 'test_hcl_value'
                }
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_xml(self):
        reply = parse_xml_file(test_files_location + "/test.xml")
        expected_reply = {
            "xml_root": {
                'file_type': 'xml',
                'test_bool_false': 'false',
                'test_bool_true': 'true',
                'test_float': '123.123',
                'test_int': '123',
                'test_xml': {
                    'test_xml_key': 'test_xml_value'
                },
                'test_list': {
                    'element': [
                        'test1',
                        'test2',
                        'test3'
                    ]
                },
                'test_string': 'testing'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_toml(self):
        reply = parse_toml_file(test_files_location + "/test.toml")
        expected_reply = {
            'file_type': "toml",
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_toml': {
                'test_toml_key': 'test_toml_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_yaml(self):
        reply = parse_yaml_file(test_files_location + "/test.yaml")
        expected_reply = {
            'file_type': "yaml",
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_yaml': {
                'test_yaml_key': 'test_yaml_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_type_estimate_string(self):
        reply = estimate_type("this_is_a_string")
        self.assertEqual(reply, "this_is_a_string")

    def test_type_estimate_string_complex(self):
        reply = estimate_type("kafka:8082")
        self.assertEqual(reply, "kafka:8082")

    def test_type_estimate_string_very_complex(self):
        reply = estimate_type("https://test//seg34\\^#%#^&@@GGH\nE#TGddvs.36230.54164:8082")
        self.assertEqual(reply, "https://test//seg34\\^#%#^&@@GGH\nE#TGddvs.36230.54164:8082")

    def test_type_estimate_false(self):
        reply_lowercase = estimate_type("false")
        reply_uppercase = estimate_type("FALSE")
        reply_mixed = estimate_type("False")
        self.assertEqual(reply_lowercase, False)
        self.assertEqual(reply_uppercase, False)
        self.assertEqual(reply_mixed, False)

    def test_type_estimate_true(self):
        reply_lowercase = estimate_type("true")
        reply_uppercase = estimate_type("TRUE")
        reply_mixed = estimate_type("True")
        self.assertEqual(reply_lowercase, True)
        self.assertEqual(reply_uppercase, True)
        self.assertEqual(reply_mixed, True)

    def test_type_estimate_int(self):
        reply = estimate_type("123")
        self.assertEqual(reply, 123)

    def test_type_estimate_float(self):
        reply = estimate_type("123.123")
        self.assertEqual(reply, 123.123)

    def test_type_estimate_list(self):
        reply = estimate_type("['test1', 123, True]")
        self.assertEqual(reply, ['test1', 123, True])
        reply = estimate_type('["test1", "test2", "test3"]')
        self.assertListEqual(reply, ["test1", "test2", "test3"])
        reply = estimate_type('["test1", {"test_key": "test_value"}, "test3", None]')
        self.assertListEqual(reply, ["test1", {"test_key": "test_value"}, "test3", None])

    def test_type_estimate_dict(self):
        reply = estimate_type("{'test_key': ['test1', 123, {'key': 'value'}]}")
        self.assertEqual(reply, {'test_key': ['test1', 123, {'key': 'value'}]})

    def test_type_estimate_none(self):
        reply_empty = estimate_type("")
        reply_lowercase = estimate_type("none")
        reply_null = estimate_type("null")
        reply_uppercase = estimate_type("NONE")
        reply_mixed = estimate_type("None")
        self.assertEqual(reply_empty, None)
        self.assertEqual(reply_lowercase, None)
        self.assertEqual(reply_null, None)
        self.assertEqual(reply_uppercase, None)
        self.assertEqual(reply_mixed, None)

    def test_type_estimate_none_custom(self):
        allowed_none_values = {"null", "none"}
        reply_empty = estimate_type("", allowed_none_values)
        reply_lowercase = estimate_type("none", allowed_none_values)
        reply_null = estimate_type("null", allowed_none_values)
        reply_uppercase = estimate_type("NONE", allowed_none_values)
        reply_mixed = estimate_type("None", allowed_none_values)
        self.assertNotEqual(reply_empty, None)
        self.assertEqual(reply_lowercase, None)
        self.assertEqual(reply_null, None)
        self.assertEqual(reply_uppercase, None)
        self.assertEqual(reply_mixed, None)

    def test_parser_init_no_recurse(self):
        expected_config_files_dict = {
            'env': [
                "test.env"
            ],
            'json': [
                'test.json',
                'test_none_values.json'
            ],
            'yaml': [
                'test.yaml'
            ],
            'yml': [],
            'toml': ['test.toml'],
            'tml': [],
            'hcl': ['test.hcl'],
            'tf': [],
            'conf': [],
            'cfg': [],
            'ini': [
                'test.ini'
            ],
            'xml': [
                'test.xml'
            ]
        }
        expected_config_type_priority = ['cli_args', 'env_vars', "env", 'json', 'yaml', 'yml', 'toml', 'tml', 'hcl',
                                         'tf', 'conf', 'cfg', 'ini', 'xml']
        parser = ParseIt(config_type_priority=None, global_default_value=None, type_estimate=True,
                         force_envvars_uppercase=True, config_location=test_files_location, envvar_prefix=None)
        self.assertDictEqual(parser.config_files_dict, expected_config_files_dict)
        self.assertEqual(parser.config_location, test_files_location)
        self.assertListEqual(parser.config_type_priority, expected_config_type_priority)
        self.assertEqual(parser.envvar_prefix, '')
        self.assertTrue(parser.force_envvars_uppercase)
        self.assertIsNone(parser.global_default_value)
        self.assertTrue(parser.type_estimate)

    def test_parser_init_recurse(self):
        expected_config_files_dict = {
            'env': [
                'test.env'
            ],
            'json': [
                'test.json',
                'test_none_values.json',
                'test_subfolder_1/test_sub_subfolder_2/test_subfolder_2.json',
                'test_subfolder_1/test_sub_subfolder_3/test_subfolder_3.json',
                'test_subfolder_1/test_sub_subfolder_3/test_subfolder_4.json',
                'test_subfolder_1/test_subfolder_1.json'
            ],
            'yaml': [
                'test.yaml'
            ],
            'yml': [],
            'toml': ['test.toml'],
            'tml': [],
            'hcl': ['test.hcl'],
            'tf': [],
            'conf': [],
            'cfg': [],
            'ini': [
                'test.ini'
            ],
            'xml': [
                'test.xml'
            ]
        }
        expected_config_type_priority = ['cli_args', 'env_vars', 'env', 'json', 'yaml', 'yml', 'toml', 'tml', 'hcl',
                                         'tf', 'conf', 'cfg', 'ini', 'xml']
        parser = ParseIt(config_type_priority=None, global_default_value=None, type_estimate=True, recurse=True,
                         force_envvars_uppercase=True, config_location=test_files_location, envvar_prefix=None)
        self.assertDictEqual(parser.config_files_dict, expected_config_files_dict)
        self.assertEqual(parser.config_location, test_files_location)
        self.assertListEqual(parser.config_type_priority, expected_config_type_priority)
        self.assertEqual(parser.envvar_prefix, '')
        self.assertTrue(parser.force_envvars_uppercase)
        self.assertIsNone(parser.global_default_value)
        self.assertTrue(parser.type_estimate)

    def test_parser_init_single_file(self):
        expected_config_files_dict = {
            'env': [],
            'cfg': [],
            'conf': [],
            'hcl': [],
            'ini': [],
            'json': [
                'test/test_files/test.json'
            ],
            'tf': [],
            'tml': [],
            'toml': [],
            'xml': [],
            'yaml': [],
            'yml': []
        }
        expected_config_type_priority = ['cli_args', 'env_vars', 'env', 'json', 'yaml', 'yml', 'toml', 'tml', 'hcl',
                                         'tf', 'conf', 'cfg', 'ini', 'xml']
        parser = ParseIt(config_type_priority=None, global_default_value=None, type_estimate=True,
                         force_envvars_uppercase=True, config_location=test_files_location + "/test.json",
                         envvar_prefix=None)
        self.assertDictEqual(parser.config_files_dict, expected_config_files_dict)
        self.assertEqual(parser.config_location, test_files_location + "/test.json")
        self.assertListEqual(parser.config_type_priority, expected_config_type_priority)
        self.assertEqual(parser.envvar_prefix, '')
        self.assertTrue(parser.force_envvars_uppercase)
        self.assertIsNone(parser.global_default_value)
        self.assertTrue(parser.type_estimate)

    def test_parser_init_envvar_divider_none(self):
        parser = ParseIt(config_location=test_files_location + "/test.json", envvar_divider=None)
        self.assertFalse(parser.nest_envvars)

    def test_parser_init_envvar_divider_set(self):
        test_envvar_divider = "_"
        parser = ParseIt(config_location=test_files_location + "/test.json", envvar_divider=test_envvar_divider)
        self.assertTrue(parser.nest_envvars)
        self.assertEqual(test_envvar_divider, parser.envvar_divider)

    def test_parser_custom_suffix_mapping_set_config_type_priority_not_set_raise_warning(self):
        with self.assertWarns(Warning):
            ParseIt(config_location=test_files_location, custom_suffix_mapping={"yaml": ["custom"]})

    def test_parser_custom_suffix_mapping_set(self):
        parser = ParseIt(config_location=test_files_location, custom_suffix_mapping={"yaml": ["custom"]},
                         config_type_priority=["custom"] + VALID_FILE_TYPE_EXTENSIONS)
        expected_config_type_priority = ["custom"] + VALID_FILE_TYPE_EXTENSIONS
        expected_valid_type_extension = ['env', 'json', 'yaml', 'yml', 'toml', 'tml', 'hcl', 'tf', 'conf', 'cfg', 'ini',
                                         'xml', 'custom']
        expected_suffix_file_type_mapping = {
            'env': [
                'env'
            ],
            'json': [
                'json'
            ],
            'yaml': [
                'yaml',
                'yml',
                'custom'
            ],
            'toml': [
                'toml',
                'tml'
            ],
            'hcl': [
                'hcl',
                'tf'
            ],
            'ini': [
                'conf',
                'cfg',
                'ini'
            ],
            'xml': [
                'xml'
            ]
        }
        reply = parser.read_configuration_variable("test_string")
        self.assertEqual("testing_custom", reply)
        reply = parser.read_configuration_variable("test_json")
        self.assertEqual({'test_json_key': 'test_json_value'}, reply)
        self.assertListEqual(expected_config_type_priority, parser.config_type_priority)
        self.assertListEqual(expected_valid_type_extension, parser.valid_file_type_extension)
        self.assertDictEqual(expected_suffix_file_type_mapping, parser.suffix_file_type_mapping)

    def test_parser_read_configuration_variable(self):
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_configuration_variable("file_type")
        self.assertEqual(reply_json, "env")
        reply_yaml = parser.read_configuration_variable("test_yaml")
        self.assertDictEqual(reply_yaml, {'test_yaml_key': 'test_yaml_value'})

    def test_parser_read_configuration_subfolder(self):
        parser = ParseIt(config_location=test_files_location, recurse=True)
        reply = parser.read_configuration_variable("test_json_subfolder")
        self.assertTrue(reply)

    def test_parser_read_configuration_recurse_false(self):
        parser = ParseIt(config_location=test_files_location, recurse=False)
        reply = parser.read_configuration_variable("test_json_subfolder")
        expected_config_files_dict = {
            'env': [
                'test.env'
            ],
            'json': [
                'test.json',
                'test_none_values.json'
            ],
            'yaml': [
                'test.yaml'
            ],
            'yml': [],
            'toml': [
                'test.toml'
            ],
            'tml': [],
            'conf': [],
            'hcl': ['test.hcl'],
            'tf': [],
            'cfg': [],
            'ini': [
                'test.ini'
            ],
            'xml': [
                'test.xml'
            ]
        }
        self.assertDictEqual(parser.config_files_dict, expected_config_files_dict)
        self.assertIsNone(reply)

    def test_parser_read_configuration_single_file(self):
        parser = ParseIt(config_location=test_files_location + "/test.hcl")
        reply = parser.read_configuration_variable("file_type")
        expected_reply = "hcl"
        self.assertEqual(reply, expected_reply)

    def test_parser_read_configuration_variable_default_value(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_configuration_variable("file_type123", default_value="test123")
        self.assertEqual(reply, "test123")

    def test_parser_read_configuration_variable_required_true_value_not_given(self):
        parser = ParseIt(config_location=test_files_location)
        with self.assertRaises(ValueError):
            parser.read_configuration_variable("file_type123", required=True)

    def test_parser_config_type_priority_wrong_type_given(self):
        with self.assertRaises(ValueError):
            parser = ParseIt(config_type_priority=['non_existing_type', 'json'])
            parser.read_configuration_variable("file_type123", required=True)

    def test_parser_read_configuration_variable_required_true_value_given_file(self):
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_configuration_variable("file_type", required=True)
        self.assertEqual(reply_json, "env")

    def test_parser_read_configuration_variable_required_true_value_given_envvar(self):
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_INT"] = "123"
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_INT", required=True)
        self.assertEqual(reply_json, 123)

    def test_parser_read_configuration_variable_global_default_value(self):
        parser = ParseIt(global_default_value="my_last_resort", config_location=test_files_location)
        reply = parser.read_configuration_variable("this_does_not_exist")
        self.assertEqual(reply, "my_last_resort")

    def test_parser_read_configuration_variable_config_type_priority(self):
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_INT"] = "123"
        parser = ParseIt(config_type_priority=["yaml", "toml", "ini", "json", "envvars", "env"],
                         config_location=test_files_location)
        reply = parser.read_configuration_variable("file_type")
        self.assertEqual(reply, "yaml")
        reply = parser.read_configuration_variable("test_toml")
        self.assertDictEqual(reply, {'test_toml_key': 'test_toml_value'})
        reply = parser.read_configuration_variable("test_ini")
        self.assertDictEqual(reply, {'test_ini_key': 'test_ini_value'})
        reply = parser.read_configuration_variable("test_json")
        self.assertDictEqual(reply, {'test_json_key': 'test_json_value'})
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_INT")
        self.assertEqual(reply, 123)
        reply = parser.read_configuration_variable("test_list")
        self.assertEqual(reply, ['test1', 'test2', 'test3'])

    def test_parser_read_configuration_variable_type_estimate_false(self):
        parser = ParseIt(type_estimate=False, config_location=test_files_location)
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_INT"] = "123"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_STRING"] = "test"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_BOOL_TRUE"] = "true"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_BOOL_FALSE"] = "false"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_LIST"] = "['test', False, 3]"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_DICT"] = "{'string': 'string', 'int': 1}"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_RECURSIVE_LIST"] = "['test', False, 3, ['test', True, 35, {'test': 12}]]"
        os.environ["TEST_ENVVAR_ESTIMATE_FALSE_RECURSIVE_DICT"] = "{'string': 'string', 'int': 1, 'dict': " \
                                                                  "{'test': [1, 2]}}"
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_INT")
        self.assertEqual(reply, "123")
        self.assertNotEqual(reply, 123)
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_STRING")
        self.assertEqual(reply, "test")
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_BOOL_TRUE")
        self.assertEqual(reply, "true")
        self.assertNotEqual(reply, True)
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_BOOL_FALSE")
        self.assertEqual(reply, "false")
        self.assertNotEqual(reply, False)
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_LIST")
        self.assertEqual(reply, "['test', False, 3]")
        self.assertNotEqual(reply, ['test', False, 3])
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_DICT")
        self.assertEqual(reply, "{'string': 'string', 'int': 1}")
        self.assertNotEqual(reply, {'string': 'string', 'int': 1})
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_RECURSIVE_LIST")
        self.assertEqual(reply, "['test', False, 3, ['test', True, 35, {'test': 12}]]")
        self.assertNotEqual(reply, ['test', False, 3, ['test', True, 35, {'test': 12}]])
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_FALSE_RECURSIVE_DICT")
        self.assertEqual(reply, "{'string': 'string', 'int': 1, 'dict': {'test': [1, 2]}}")
        self.assertNotEqual(reply, {'string': 'string', 'int': 1, 'dict': {'test': [1, 2]}})

    def test_parser_read_configuration_variable_type_estimate_true(self):
        parser = ParseIt(type_estimate=True, config_location=test_files_location)
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_INT"] = "123"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_STRING"] = "test"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_BOOL_TRUE"] = "true"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_BOOL_FALSE"] = "false"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_LIST"] = "['test', False, 3]"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_DICT"] = "{'string': 'string', 'int': 1}"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_RECURSIVE_LIST"] = "['test', False, 3, ['test', True, 35, {'test': 12}]]"
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_RECURSIVE_DICT"] = "{'string': 'string', 'int': 1, 'dict': " \
                                                                 "{'test': [1, 2]}}"
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_INT")
        self.assertNotEqual(reply, "123")
        self.assertEqual(reply, 123)
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_STRING")
        self.assertEqual(reply, "test")
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_BOOL_TRUE")
        self.assertNotEqual(reply, "true")
        self.assertEqual(reply, True)
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_BOOL_FALSE")
        self.assertNotEqual(reply, "false")
        self.assertEqual(reply, False)
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_LIST")
        self.assertNotEqual(reply, "['test', False, 3]")
        self.assertListEqual(reply, ['test', False, 3])
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_DICT")
        self.assertNotEqual(reply, "{'string': 'string', 'int': 1}")
        self.assertDictEqual(reply, {'string': 'string', 'int': 1})
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_RECURSIVE_LIST")
        self.assertNotEqual(reply, "['test', False, 3, ['test', True, 35, {'test': 12}]]")
        self.assertListEqual(reply, ['test', False, 3, ['test', True, 35, {'test': 12}]])
        reply = parser.read_configuration_variable("TEST_ENVVAR_ESTIMATE_TRUE_RECURSIVE_DICT")
        self.assertNotEqual(reply, "{'string': 'string', 'int': 1, 'dict': {'test': [1, 2]}}")
        self.assertDictEqual(reply, {'string': 'string', 'int': 1, 'dict': {'test': [1, 2]}})

    def test_parser_read_configuration_variable_force_envvars_uppercase_true(self):
        parser = ParseIt(force_envvars_uppercase=True)
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_INT"] = "123"
        os.environ["test_envvar_estimate_true_int"] = "456"
        reply = parser.read_configuration_variable("test_envvar_estimate_true_int")
        self.assertEqual(reply, 123)
        self.assertNotEqual(reply, 456)

    def test_parser_read_configuration_variable_force_envvars_uppercase_false(self):
        parser = ParseIt(force_envvars_uppercase=False, config_location=test_files_location)
        os.environ["TEST_ENVVAR_ESTIMATE_TRUE_INT"] = "123"
        os.environ["test_envvar_estimate_true_int"] = "456"
        reply = parser.read_configuration_variable("test_envvar_estimate_true_int")
        self.assertNotEqual(reply, 123)
        self.assertEqual(reply, 456)

    def test_parser_read_configuration_variable_config_location(self):
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_configuration_variable("file_type")
        self.assertEqual(reply_json, "env")
        reply_json = parser.read_configuration_variable("test_float")
        self.assertEqual(reply_json, 123.123)
        reply_yaml = parser.read_configuration_variable("test_yaml")
        self.assertDictEqual(reply_yaml, {'test_yaml_key': 'test_yaml_value'})
        reply_xml = parser.read_configuration_variable("xml_root")
        expected_reply_xml = {
            'file_type': 'xml',
            'test_bool_false': False,
            'test_bool_true': True,
            'test_float': 123.123,
            'test_int': 123,
            'test_xml': {
                'test_xml_key': 'test_xml_value'
            },
            'test_list': {
                'element': [
                    'test1',
                    'test2',
                    'test3'
                ]
            },
            'test_string': 'testing'
        }
        self.assertDictEqual(reply_xml, expected_reply_xml)
        reply_hcl = parser.read_configuration_variable("test_hcl")
        expected_reply_hcl = {
            'test_hcl_name': {
                'test_hcl_key': 'test_hcl_value'
            }
        }
        self.assertDictEqual(reply_hcl, expected_reply_hcl)

    def test_parser_read_configuration_variable_envvar_prefix(self):
        parser = ParseIt(envvar_prefix="prefix_test_", config_location=test_files_location)
        os.environ["PREFIX_TEST_TEST_ENVVAR_ESTIMATE_TRUE_INT"] = "123"
        reply = parser.read_configuration_variable("test_envvar_estimate_true_int")
        self.assertEqual(reply, 123)

    def test_parser_read_configuration_variable_envvar_nested(self):
        parser = ParseIt(config_location=test_files_location, envvar_divider="_")
        os.environ["TEST_ENVVAR_NESTED"] = "123"
        reply = parser.read_configuration_variable("test_envvar_nested")
        self.assertDictEqual(reply, {"test": {"envvar": {"nested": 123}}})

    def test_envvar_defined_true(self):
        os.environ["TEST_ENV"] = "123"
        reply = envvar_defined("TEST_ENV")
        self.assertTrue(reply)

    def test_envvar_defined_false(self):
        reply = envvar_defined("TEST_ENV")
        self.assertFalse(reply)

    def test_envvar_defined_true_upper_case(self):
        os.environ["TEST_ENV"] = "123"
        reply = envvar_defined("test_env", force_uppercase=True)
        self.assertTrue(reply)

    def test_envvar_defined_false_upper_case(self):
        os.environ["TEST_ENV"] = "123"
        reply = envvar_defined("test_env", force_uppercase=False)
        self.assertFalse(reply)

    def test_read_all_envvars_to_dict_force_uppercase_true(self):
        test_envvars = {"TEST_ENV": "123", "test_env_lowercase": "456"}
        with mock.patch.dict(os.environ, test_envvars):
            reply = read_all_envvars_to_dict(force_uppercase=True)
            self.assertEqual(type(reply), dict)
            self.assertEqual(reply["test_env"], "123")
            self.assertEqual(reply["test_env_lowercase"], "456")

    def test_read_all_envvars_to_dict_force_uppercase_false(self):
        test_envvars = {"TEST_ENV": "123", "test_env_lowercase": "456"}
        with mock.patch.dict(os.environ, test_envvars):
            reply = read_all_envvars_to_dict(force_uppercase=False)
            self.assertEqual(type(reply), dict)
            self.assertEqual(reply["TEST_ENV"], "123")
            self.assertEqual(reply["test_env_lowercase"], "456")

    def test_envvars_split_force_uppercase_false(self):
        expected_reply = {'dict': {'subdict': {'key': 'example_value'}}}
        test_key = "dict_subdict_key"
        test_value = "example_value"
        reply = split_envvar(test_key, test_value, divider="_")
        self.assertEqual(expected_reply, reply)

    def test_envvars_split_force_uppercase_true(self):
        expected_reply = {'dict': {'subdict': {'key': 'example_value'}}}
        test_key = "dict_subdict_key"
        test_value = "example_value"
        reply = split_envvar(test_key, test_value, divider="_")
        self.assertEqual(expected_reply, reply)

    def test_envvars_split_no_nesting_needed(self):
        expected_reply = {'dict': 'example_value'}
        test_key = "dict"
        test_value = "example_value"
        reply = split_envvar(test_key, test_value, divider="_")
        self.assertEqual(expected_reply, reply)

    def test_envvars_split_custom_divider(self):
        expected_reply = {'dict': {'subdict': {'key': 'example_value'}}}
        test_key = "dict.subdict.key"
        test_value = "example_value"
        reply = split_envvar(test_key, test_value, divider=".")
        self.assertEqual(expected_reply, reply)

    def test_envvars_split_envvar_combained_dict(self):
        test_envvars = {"TEST_SPLIT_UPPERCASE1": "test123", "test_split_lowercase1": "test456"}
        with mock.patch.dict(os.environ, test_envvars):
            reply = split_envvar_combained_dict()
            self.assertEqual(type(reply), dict)
            self.assertEqual(reply["test"]["split"]["uppercase1"], "test123")
            self.assertEqual(reply["test"]["split"]["lowercase1"], "test456")

    def test_envvars_split_envvar_combained_dict_custom_divider(self):
        test_envvars = {"TEST.SPLIT.UPPER_CASE2": "test123", "test.split.lower_case2": "test456"}
        with mock.patch.dict(os.environ, test_envvars):
            reply = split_envvar_combained_dict(divider=".")
            self.assertEqual(type(reply), dict)
            self.assertEqual(reply["test"]["split"]["upper_case2"], "test123")
            self.assertEqual(reply["test"]["split"]["lower_case2"], "test456")

    def test_envvars_split_envvar_combained_dict_force_uppercase_false(self):
        test_envvars = {
            "TEST_SPLIT_UPPERCASE3": "test123",
            "test_split_lowercase3": "test456",
            "test_split_lowercase4": "test789"
        }
        with mock.patch.dict(os.environ, test_envvars):
            reply = split_envvar_combained_dict(force_uppercase=False)
            self.assertEqual(type(reply), dict)
            self.assertEqual(reply["TEST"]["SPLIT"]["UPPERCASE3"], "test123")
            self.assertEqual(reply["test"]["split"]["lowercase3"], "test456")
            self.assertEqual(reply["test"]["split"]["lowercase4"], "test789")

    def test_parser_config_found_in_key(self):
        parser = ParseIt(config_location=test_files_location)
        config_found_reply, config_value_reply = parser._check_config_in_dict("test_key", {"test_key": "test_value"})
        self.assertTrue(config_found_reply)
        self.assertEqual(config_value_reply, "test_value")

    def test_parser_config_found_in_key_false(self):
        parser = ParseIt(config_location=test_files_location)
        config_found_reply, config_value_reply = parser._check_config_in_dict("wrong_key", {"test_key": "test_value"})
        self.assertFalse(config_found_reply)
        self.assertEqual(config_value_reply, None)

    def test_parser__parse_file_per_type_wrong_type(self):
        parser = ParseIt(config_location=test_files_location)
        with self.assertRaises(ValueError):
            parser._parse_file_per_type("non_existing_type", test_files_location + "/test.json")

    def test_parser__parse_file_per_type_json(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser._parse_file_per_type("json", test_files_location + "/test.json")
        expected_reply = {
            'file_type': 'json',
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_json': {
                'test_json_key': 'test_json_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_parser__parse_file_per_type_yaml(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser._parse_file_per_type("yaml", test_files_location + "/test.yaml")
        expected_reply = {
            'file_type': 'yaml',
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_yaml': {
                'test_yaml_key': 'test_yaml_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)
        reply = parser._parse_file_per_type("yml", test_files_location + "/test.yaml")
        self.assertDictEqual(reply, expected_reply)

    def test_parser__parse_file_per_type_custom_yaml(self):
        parser = ParseIt(config_location=test_files_location, custom_suffix_mapping={"yaml": ["custom"]},
                         config_type_priority=["custom"] + VALID_FILE_TYPE_EXTENSIONS)
        reply = parser._parse_file_per_type("custom", test_files_location + "/test.custom")
        expected_reply = {
            'file_type': 'custom_yaml_suffix',
            'test_string': 'testing_custom',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_yaml': {
                'test_yaml_key': 'custom_test_yaml_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_parser__parse_file_per_type_toml(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser._parse_file_per_type("toml", test_files_location + "/test.toml")
        expected_reply = {
            'file_type': 'toml',
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_toml': {
                'test_toml_key': 'test_toml_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)
        reply = parser._parse_file_per_type("tml", test_files_location + "/test.toml")
        self.assertDictEqual(reply, expected_reply)

    def test_parser__parse_file_per_type_hcl(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser._parse_file_per_type("hcl", test_files_location + "/test.hcl")
        expected_reply = {
            'file_type': 'hcl',
            'test_string': 'testing',
            'test_bool_true': True,
            'test_bool_false': False,
            'test_int': 123,
            'test_float': 123.123,
            'test_dict': {
                "hcl_dict_key": "hcl_dict_value"
            },
            'test_list': [
                'test1',
                'test2',
                'test3'
            ],
            'test_hcl': {
                "test_hcl_name": {
                    'test_hcl_key': 'test_hcl_value'
                }
            }
        }
        self.assertDictEqual(reply, expected_reply)
        reply = parser._parse_file_per_type("tf", test_files_location + "/test.hcl")
        self.assertDictEqual(reply, expected_reply)

    def test_parser__parse_file_per_type_ini(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser._parse_file_per_type("ini", test_files_location + "/test.ini")
        expected_reply = {
            'DEFAULT': {
                'file_type': 'ini',
                'test_string': 'testing',
                'test_bool_true': 'true',
                'test_bool_false': 'false',
                'test_int': '123.0',
                'test_float': '123.123',
                'test_list': '["test1", "test2", "test3"]'
            },
            'test_ini': {
                'test_ini_key': 'test_ini_value'
            }
        }
        self.assertDictEqual(reply, expected_reply)
        reply = parser._parse_file_per_type("conf", test_files_location + "/test.ini")
        self.assertDictEqual(reply, expected_reply)
        reply = parser._parse_file_per_type("cfg", test_files_location + "/test.ini")
        self.assertDictEqual(reply, expected_reply)

    def test_parser__parse_file_per_type_xml(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser._parse_file_per_type("xml", test_files_location + "/test.xml")
        expected_reply = {
            'xml_root': {
                'file_type': 'xml',
                'test_bool_false': 'false',
                'test_bool_true': 'true',
                'test_float': '123.123',
                'test_int': '123',
                'test_xml': {
                    'test_xml_key': 'test_xml_value'
                },
                'test_list': {
                    'element': [
                        'test1',
                        'test2',
                        'test3'
                    ]
                },
                'test_string': 'testing'
            }
        }
        self.assertDictEqual(reply, expected_reply)

    def test_command_line_args_read_command_line_arg(self):
        test_args = ["parse_it_mock_script.py", "--test_key", "test_value"]
        with mock.patch('sys.argv', test_args):
            reply = read_command_line_arg("test_key")
            self.assertEqual(reply, "test_value")

    def test_command_line_args_read_command_line_arg_not_defined(self):
        test_args = ["parse_it_mock_script.py", "--test_key", "test_value"]
        with mock.patch('sys.argv', test_args):
            reply = read_command_line_arg("non_existing_test_key")
            self.assertIsNone(reply)

    def test_command_line_args_command_line_arg_defined_false(self):
        test_args = ["parse_it_mock_script.py", "--test_key", "test_value"]
        with mock.patch('sys.argv', test_args):
            reply = command_line_arg_defined("non_existing_test_key")
            self.assertFalse(reply)

    def test_command_line_args_command_line_arg_defined_true(self):
        test_args = ["parse_it_mock_script.py", "--test_key", "test_value"]
        with mock.patch('sys.argv', test_args):
            reply = command_line_arg_defined("test_key")
            self.assertTrue(reply)

    def test_command_line_args_read_all_cli_args_to_dict(self):
        test_args = ["parse_it_mock_script.py", "--test_key", "test_value", "--another_test", "another_value"]
        expected_reply = {"test_key": "test_value", "another_test": "another_value"}
        with mock.patch('sys.argv', test_args):
            reply = read_all_cli_args_to_dict()
            self.assertEqual(reply, expected_reply)

    def test_command_line_args_read_all_cli_args_to_dict_no_cli_args(self):
        test_args = ["parse_it_mock_script.py"]
        expected_reply = {}
        with mock.patch('sys.argv', test_args):
            reply = read_all_cli_args_to_dict()
            self.assertEqual(reply, expected_reply)

    def test_command_line_args_read_all_cli_args_to_dict_no_cli_args_with_double_dash(self):
        test_args = ["parse_it_mock_script.py", "-wrong_format_test_key", "test_value"]
        expected_reply = {}
        with mock.patch('sys.argv', test_args):
            reply = read_all_cli_args_to_dict()
            self.assertEqual(reply, expected_reply)

    def test_parser_read_configuration_from_cli_arg(self):
        test_args = ["parse_it_mock_script.py", "--test_cli_key", "test_value", "--test_cli_int", "123"]
        with mock.patch('sys.argv', test_args):
            parser = ParseIt()
            reply = parser.read_configuration_variable("test_cli_key")
            self.assertEqual(reply, "test_value")
            reply = parser.read_configuration_variable("test_cli_int")
            self.assertEqual(reply, 123)

    def test_parser_read_configuration_variable_check_allowed_types_None(self):
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_configuration_variable("file_type", allowed_types=None)
        self.assertEqual(reply_json, "env")

    def test_parser_read_configuration_variable_check_allowed_not_in_types_list(self):
        with self.assertRaises(TypeError):
            parser = ParseIt(config_location=test_files_location)
            parser.read_configuration_variable("file_type", allowed_types=[int, list, dict, None, float])

    def test_parser_read_configuration_variable_check_allowed_in_types_list(self):
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_configuration_variable("file_type", allowed_types=[int, str, dict, None, float])
        self.assertEqual(reply_json, "env")

    def test_parser_read_multiple_configuration_variables_works_multiple_list(self):
        expected_reply = {
            'file_type': 'env',
            'test_hcl': {
                'test_hcl_name': {
                    'test_hcl_key': 'test_hcl_value'
                }
            },
            'test_int': 123,
            'test_list': [
                'test1',
                'test2',
                'test3'
            ]
        }
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_multiple_configuration_variables(["file_type", "test_hcl", "test_int", "test_list"])
        self.assertDictEqual(reply_json, expected_reply)

    def test_parser_read_multiple_configuration_variables_default_value(self):
        expected_reply = {
            'file_type': 'env',
            'non_existing_value': 'test'
        }
        parser = ParseIt(config_location=test_files_location)
        reply_json = parser.read_multiple_configuration_variables(["file_type", "non_existing_value"],
                                                                  default_value="test")
        self.assertDictEqual(reply_json, expected_reply)

    def test_parser_read_multiple_configuration_variables_required_true_raise_error(self):
        parser = ParseIt(config_location=test_files_location)
        with self.assertRaises(ValueError):
            parser.read_multiple_configuration_variables(["file_type", "non_existing_value"], required=True)

    def test_parser_read_multiple_configuration_variables_allowed_types_raise_error(self):
        parser = ParseIt(config_location=test_files_location)
        with self.assertRaises(TypeError):
            parser.read_multiple_configuration_variables(["file_type", "non_existing_value"], allowed_types=[int, dict])

    def test_parser_read_all_configuration_variables(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables()
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_nest_envvars(self):
        parser = ParseIt(config_location=test_files_location, envvar_divider="_", force_envvars_uppercase=True)
        test_envvars = {"test_split_nesting1": "test123", "test_split_nesting2": "test456"}
        with mock.patch.dict(os.environ, test_envvars):
            reply = parser.read_all_configuration_variables()
            self.assertDictEqual(reply["test"]["split"], {'nesting1': 'test123', 'nesting2': 'test456'})
            self.assertEqual(reply["file_type"], "env")
            self.assertEqual(reply["test_string"], "testing")
            self.assertTrue(reply["test_bool_true"])
            self.assertFalse(reply["test_bool_false"])
            self.assertEqual(reply["test_int"], 123)
            self.assertEqual(reply["test_float"], 123.123)
            self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
            self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_default_value_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(default_value={"default_value_test": "it_works"})
        self.assertEqual(reply["default_value_test"], "it_works")
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_none_default_value_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(default_value=None)
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_empty_default_value_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(default_value={})
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_required_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(required=["file_type", "test_int"])
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_empty_required_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(required=[])
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_none_required_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(required=None)
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_allowed_types_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(allowed_types={
            "file_type": [
                str,
                bool
            ],
            "test_int": [
                int,
                str
            ]
        })
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_empty_allowed_types_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(allowed_types={})
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_none_allowed_types_given(self):
        parser = ParseIt(config_location=test_files_location)
        reply = parser.read_all_configuration_variables(allowed_types=None)
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_type_estimate_true(self):
        test_envvars = {"TEST_ENV_TYPE_ESTIMATE": "123"}
        with mock.patch.dict(os.environ, test_envvars):
            parser = ParseIt(config_location=test_files_location, type_estimate=True)
            reply = parser.read_all_configuration_variables()
            self.assertEqual(reply["test_env_type_estimate"], 123)

    def test_parser_read_all_configuration_variables_type_estimate_false(self):
        test_envvars = {"TEST_ENV_TYPE_ESTIMATE": "123"}
        with mock.patch.dict(os.environ, test_envvars):
            parser = ParseIt(config_location=test_files_location, type_estimate=False)
            reply = parser.read_all_configuration_variables()
            self.assertEqual(reply["test_env_type_estimate"], "123")

    def test_parser_read_all_configuration_variables_raise_allowed_types_error(self):
        parser = ParseIt(config_location=test_files_location)
        with self.assertRaises(TypeError):
            parser.read_all_configuration_variables(allowed_types={"file_type": [bool, dict]})

    def test_parser_read_all_configuration_variables_raise_required_error(self):
        parser = ParseIt(config_location=test_files_location)
        with self.assertRaises(ValueError):
            parser.read_all_configuration_variables(required=["non_existing_key"])

    def test_parser_read_all_configuration_variables_config_type_in_data_sources_error(self):
        parser = ParseIt(config_type_priority=['non_existing_type', 'json'])
        with self.assertRaises(ValueError):
            parser.read_all_configuration_variables(required=["non_existing_key"])

    def test_parser_read_all_configuration_variables_read_envvars(self):
        test_envvars = {"TEST_ENV_TYPE_ESTIMATE": "123"}
        with mock.patch.dict(os.environ, test_envvars):
            parser = ParseIt(config_location=test_files_location, type_estimate=True)
            reply = parser.read_all_configuration_variables()
            self.assertEqual(reply["test_env_type_estimate"], 123)
            self.assertEqual(reply["file_type"], "env")
            self.assertEqual(reply["test_string"], "testing")
            self.assertTrue(reply["test_bool_true"])
            self.assertFalse(reply["test_bool_false"])
            self.assertEqual(reply["test_int"], 123)
            self.assertEqual(reply["test_float"], 123.123)
            self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
            self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_read_cli_args(self):
        test_args = ["parse_it_mock_script.py", "--test_cli_key_no_folder", "test_value"]
        with mock.patch('sys.argv', test_args):
            parser = ParseIt(config_location=test_files_location, type_estimate=True)
            reply = parser.read_all_configuration_variables()
            self.assertEqual(reply["test_cli_key_no_folder"], "test_value")
            self.assertEqual(reply["file_type"], "env")
            self.assertEqual(reply["test_string"], "testing")
            self.assertTrue(reply["test_bool_true"])
            self.assertFalse(reply["test_bool_false"])
            self.assertEqual(reply["test_int"], 123)
            self.assertEqual(reply["test_float"], 123.123)
            self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
            self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_read_single_file(self):
        parser = ParseIt(config_location=test_files_location + "/test.hcl", type_estimate=True)
        reply = parser.read_all_configuration_variables()
        self.assertEqual(reply["file_type"], "hcl")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_recursive_false(self):
        parser = ParseIt(config_location=test_files_location, recurse=False)
        reply = parser.read_all_configuration_variables()
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_recursive_true(self):
        parser = ParseIt(config_location=test_files_location, recurse=True)
        reply = parser.read_all_configuration_variables()
        self.assertEqual(reply["file_type"], "env")
        self.assertEqual(reply["test_string"], "testing")
        self.assertTrue(reply["test_bool_true"])
        self.assertTrue(reply["test_json_subfolder"])
        self.assertFalse(reply["test_bool_false"])
        self.assertEqual(reply["test_int"], 123)
        self.assertEqual(reply["test_float"], 123.123)
        self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
        self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])

    def test_parser_read_all_configuration_variables_check_order(self):
        parser = ParseIt(config_location=test_files_location, config_type_priority=["env_vars", "hcl", "json"])
        test_envvars = {"TEST_INT": "12345"}
        with mock.patch.dict(os.environ, test_envvars):
            reply = parser.read_all_configuration_variables()
            self.assertEqual(reply["file_type"], "hcl")
            self.assertEqual(reply["test_string"], "testing")
            self.assertTrue(reply["test_bool_true"])
            self.assertFalse(reply["test_bool_false"])
            self.assertEqual(reply["test_int"], 12345)
            self.assertEqual(reply["test_float"], 123.123)
            self.assertDictEqual(reply["test_dict"], {'hcl_dict_key': 'hcl_dict_value'})
            self.assertListEqual(reply["test_list"], ['test1', 'test2', 'test3'])
            self.assertDictEqual(reply["test_json"], {"test_json_key": "test_json_value"})

    def test_parser_type_estimate_none(self):
        parser = ParseIt(
            config_location=f"{test_files_location}/test_none_values.json")
        reply = parser.read_all_configuration_variables()
        self.assertEqual(reply["empty_string"], None)
        self.assertEqual(reply["none_lower"], None)
        self.assertEqual(reply["none_upper"], None)
        self.assertEqual(reply["none_mixed"], None)
        self.assertEqual(reply["null"], None)

    def test_parser_type_estimate_none_custom(self):
        parser = ParseIt(
            config_location=f"{test_files_location}/test_none_values.json",
            none_values={"null", "none"})
        reply = parser.read_all_configuration_variables()
        self.assertNotEqual(reply["empty_string"], None)
        self.assertEqual(reply["none_lower"], None)
        self.assertEqual(reply["none_upper"], None)
        self.assertEqual(reply["none_mixed"], None)
        self.assertEqual(reply["null"], None)



from parse_it.file.file_reader import *
import toml


def parse_toml_file(path_to_toml_file: str) -> dict:
    """take a path to a TOML file & returns it as a valid python dict.

            Arguments:
                path_to_toml_file -- the path of the toml file
            Returns:
                config_file_dict -- dict of the file
    """
    return toml.loads(read_file(path_to_toml_file))



from parse_it.file.file_reader import *
import json


def parse_json_file(path_to_json_file: str) -> dict:
    """take a path to a JSON file & returns it as a valid python dict.

            Arguments:
                path_to_json_file -- the path of the json file
            Returns:
                config_file_dict -- dict of the file
    """
    return json.loads(read_file(path_to_json_file))



from pathlib import Path
from typing import Optional
import os
import warnings


def read_file(file_path: str) -> Optional[str]:
    """Read a file and returns it's contents (as a string) or None if file does not exist.

     Arguments:
        file_path -- the path of the file to be read
    Returns:
        file_contents -- string of the file contents
    """
    try:
        with open(file_path) as f:
            file_contents = f.read()
        return file_contents
    except FileNotFoundError:
        return None


def folder_exists(folder_path: str) -> bool:
    """Returns True if folder exists, False otherwise.

     Arguments:
        folder_path -- the path of the folder to be checked
    Returns:
        True if folder is dir, False otherwise
    """
    possible_folder = Path(strip_trailing_slash(folder_path))
    return possible_folder.is_dir()


def file_or_folder(checked_path: str) -> Optional[str]:
    """Returns "file" if the file_path is a file, "folder" if it's a dir & None otherwise.

     Arguments:
        checked_path -- the path of the file to be checked if it's a file or a folder
    Returns:
        "folder" if the path is a folder, "file" if it's a file and None otherwise
    """
    checked_file_path = Path(strip_trailing_slash(checked_path))
    if checked_file_path.is_dir() is True:
        return "folder"
    elif checked_file_path.is_file() is True:
        return "file"
    else:
        return None


def strip_trailing_slash(folder_path: str) -> str:
    """if a folder_path ends in a slash strip it & return the path, otherwise just return the path, only edge case is
        the root folder (/) which is kept the same.

     Arguments:
        folder_path -- the path of the folder to be checked
    Returns:
        A trailing slash stripped string of folder_path
    """
    if len(folder_path) > 1 and folder_path.endswith(("/", "\\")):
        folder_path = folder_path[:-1]
    return folder_path


def file_types_in_folder(folder_path: str, file_types_endings: list, recurse: bool = True) -> dict:
    """list all the config file types found inside the given folder based on the filename extension

        Arguments:
            folder_path -- the path of the folder to be checked
            file_types_endings -- list of file types to look for
            recurse -- if True (default) will also look in all subfolders
        Returns:
            config_files_dict -- dict of {file_type: [list_of_file_names_of_said_type]}
    """
    folder_path = strip_trailing_slash(folder_path)
    if folder_exists(folder_path) is False:
        warnings.warn("config_location " + folder_path + " does not exist, only envvars & cli args will be used")
        config_files_dict = {}
        for file_type_ending in file_types_endings:
            config_files_dict[file_type_ending] = []
    else:
        config_files_dict = {}
        for file_type_ending in file_types_endings:
            config_files_dict[file_type_ending] = []
            if recurse is True:
                for root, subFolders, files in os.walk(folder_path, topdown=True):
                    for file in files:
                        if file.endswith("." + file_type_ending):
                            if folder_path + "/" in root:
                                original_root_path = root
                                if root[0] != "/":
                                    root = root.split(folder_path + "/", 1)[1]
                                config_files_dict[file_type_ending].append(os.path.join(root, file))
                                if original_root_path != root:
                                    root = original_root_path
                            else:
                                config_files_dict[file_type_ending].append(file)
            else:
                for file in os.listdir(folder_path):
                    if os.path.isfile(os.path.join(folder_path, file)) and file.endswith(file_type_ending):
                        config_files_dict[file_type_ending].append(file)
            config_files_dict[file_type_ending].sort()
    return config_files_dict



from parse_it.file.file_reader import *
import yaml


def parse_yaml_file(path_to_yaml_file: str) -> dict:
    """take a path to a YAML file & returns it as a valid python dict.

            Arguments:
                path_to_yaml_file -- the path of the yaml file
            Returns:
                config_file_dict -- dict of the file
    """
    return yaml.load(read_file(path_to_yaml_file), Loader=yaml.FullLoader)



import ast
from typing import Any
from contextlib import suppress


def estimate_type(node: Any, none_values=None) -> Any:
    """ Takes any type and return it's value in a type it estimates it to be based on ast.literal_eval & internal logic,
    if the result is a list or a dict will recurse to run all internal values as well, in case of problems parsing the
    string with ast.literal_eval it will fallback to sticking with the original type
     
            Arguments:
                node -- the string a type estimation is needed for
                none_values -- the values that should be converted to `None`
            Returns:
                node -- the value of the string in the estimated type

    """

    if none_values is None:
        none_values = {"", "null", "none"}

    # this is to support XML type estimation as it returns a dict of all strings
    if isinstance(node, dict):
        node = str(node)

    if isinstance(node, str):
        if node.lower() in {"true", "false"}:
            node = node.title()
        elif node.lower() in none_values:
            node = "None"

        with suppress(ValueError, SyntaxError):
            node = ast.literal_eval(node)
            if isinstance(node, list):
                node = [
                    estimate_type(item, none_values=none_values)
                    for item in node
                ]
            if isinstance(node, dict):
                node = {
                    key: estimate_type(value, none_values=none_values)
                    for key, value in node.items()
                }

    return node



from parse_it.command_line_args.command_line_args import *
from parse_it.envvars.envvars import *
from parse_it.file.yaml import *
from parse_it.file.ini import *
from parse_it.file.json import *
from parse_it.file.env import *
from parse_it.file.hcl import *
from parse_it.file.toml import *
from parse_it.file.xml import *
from parse_it.type_estimate.type_estimate import *
from parse_it.file.file_reader import *
from typing import Any, Tuple, Optional
import warnings


class ParseIt:

    def __init__(self, config_type_priority: Optional[list] = None, global_default_value: Any = None,
                 type_estimate: bool = True, recurse: bool = False, force_envvars_uppercase: bool = True,
                 config_location: Optional[str] = None, envvar_prefix: Optional[str] = None,
                 custom_suffix_mapping: Optional[dict] = None, envvar_divider: Optional[str] = None,
                 none_values: Optional[set] = None):
        """configures the object which is used to query all types of configuration inputs available and prioritize them
                based on your needs

                    Arguments:
                        config_type_priority -- a list of file types extensions your willing to accept, list order
                            dictates priority of said file types, default list order is as follow:
                            [ "cli_args", "envvars", "env", "json", "yaml", "yml", "toml", "tml", "hcl", "tf", "conf",
                            "cfg", "ini", "xml" ]
                            in the case of multiple files of same type they are all read and the first one that has the
                            needed key is the one used.
                            if no value is returned then the default_value declared at the read_configuration_variable
                            will be used and if that is not configured then the global_default_value will be used
                        global_default_value -- defaults to None, see config_type_priority for it's use
                        type_estimate -- if set to True (True by default) will try to automatically figure out the type
                            of the returned value on it's own, useful for envvars & ini type files which always return a
                            string otherwise
                        recurse -- False by default, if set to True will also look in all subfolders
                        force_envvars_uppercase -- if set to True (which is the default) will force all envvars keys to
                            be UPPERCASE
                        config_location -- the location where the configuration file(s) will be looked for, if None
                            (default) will look in the current working directory, accepts either a directory which will
                            then look in all files in that directory (possibly recursivly depanding on the recurse flag)
                            or if a standard file will look only in said file
                        envvar_prefix -- will add the given prefix for all envvars if set
                        custom_suffix_mapping -- a custom dict which will can map custom file suffixes to a file type
                        envvar_divider -- the divider to split an envvar to nested dicts, if set to None (default) said
                            nesting is disabled
                        none_values -- A tuple with values that should be considered `None`, if set to None (default) it
                            will take the default set tuple of {"", "null", "none"}
        """

        # first we describe the standard file type suffix mapping and what file types are are standard file extensions
        self.suffix_file_type_mapping = {
            "env": [
                "env"
            ],
            "json": [
                "json"
            ],
            "yaml": [
                "yaml",
                "yml"
            ],
            "toml": [
                "toml",
                "tml"
            ],
            "hcl": [
                "hcl",
                "tf"
            ],
            "ini": [
                "conf",
                "cfg",
                "ini"
            ],
            "xml": [
                "xml"
            ]
        }
        self.valid_file_type_extension = [
            "env",
            "json",
            "yaml",
            "yml",
            "toml",
            "tml",
            "hcl",
            "tf",
            "conf",
            "cfg",
            "ini",
            "xml"
        ]

        if envvar_divider is None:
            self.nest_envvars = False
        else:
            self.nest_envvars = True
            self.envvar_divider = envvar_divider

            # now we add any custom file suffixes the user wanted to the list of possible file extensions and valid
        # suffixes
        if custom_suffix_mapping is not None:
            for file_type, custom_file_suffix in custom_suffix_mapping.items():
                self.suffix_file_type_mapping[file_type] = self.suffix_file_type_mapping[file_type] + custom_file_suffix
            custom_suffixes_list = [suffix_value for suffix_list in custom_suffix_mapping.values() for suffix_value in
                                    suffix_list]
            self.valid_file_type_extension += custom_suffixes_list
            if config_type_priority is None:
                warnings.warn("custom_suffix_mapping is defined but config_type_priority is using the default setting, "
                              "custom file suffixes will not be used")

        if envvar_prefix is None:
            self.envvar_prefix = ""
        else:
            self.envvar_prefix = envvar_prefix

        self.global_default_value = global_default_value
        self.type_estimate = type_estimate
        self.force_envvars_uppercase = force_envvars_uppercase

        if config_type_priority is None:
            self.config_type_priority = [
                "cli_args",
                "env_vars",
                "env",
                "json",
                "yaml",
                "yml",
                "toml",
                "tml",
                "hcl",
                "tf",
                "conf",
                "cfg",
                "ini",
                "xml"
            ]
        else:
            self.config_type_priority = config_type_priority

        if config_location is None:
            self.config_location = os.getcwd()
        else:
            self.config_location = config_location

        # create a list of all valid types
        valid_config_types = []
        for config_type in self.config_type_priority:
            if config_type in self.valid_file_type_extension:
                valid_config_types.append(config_type)

        # we check if the config is a file or a folder
        self.config_file_type = file_or_folder(self.config_location)

        # if config is dict or not declared populate the config_files_dict with the list of locations for each file type
        if self.config_file_type == "folder" or self.config_file_type is None:
            self.config_files_dict = file_types_in_folder(self.config_location, valid_config_types,
                                                          recurse=recurse)
        # if the config is a file populate the config_files_dict with that single file and have the rest file types
        # be blank
        elif self.config_file_type == "file":
            self.config_files_dict = {}
            for file_type_ending in valid_config_types:
                self.config_files_dict[file_type_ending] = []
                if self.config_location.endswith(file_type_ending):
                    self.config_files_dict[file_type_ending].append(self.config_location)

        if none_values is None:
            self.none_values = {"", "null", "none"}
        else:
            self.none_values = none_values

    def read_configuration_variable(self, config_name: str, default_value: Any = None, required: bool = False,
                                    allowed_types: Optional[list] = None) -> Any:
        """reads a single key of the configuration and returns the first value of it found based on the priority of each
                config file option given in the __init__ of the class

                    Arguments:
                        config_name -- the configuration key name you want to get the value of
                        default_value -- defaults to None, see config_type_priority in class init for it's use
                        required -- defaults to False, if set to True will ignore default_value & global_default_value
                            and will raise an ValueError if the configuration is not configured in any of the config
                            files/envvars/cli args
                        allowed_types -- Defaults to None, an optional list of types that are accepted for the variable
                            to be, if set a check will be preformed and if the variables value given is not of any of
                            the types in said list a TypeError will be raised
                    Returns:
                        config_value -- the value of the configuration requested
        """

        # we start with both key not found and the value being None
        config_value = None
        config_key_found = False

        # we now loop over all the permitted types of where the config key might be and break on the first one found
        # after setting config_key_found to True and config_value to the value found
        for config_type in self.config_type_priority:
            if config_type == "cli_args":
                config_key_found = command_line_arg_defined(config_name)
                if config_key_found is True:
                    config_value = read_command_line_arg(config_name)
                    break
            elif config_type == "envvars" or config_type == "env_vars":
                config_key_found = envvar_defined(self.envvar_prefix + config_name,
                                                  force_uppercase=self.force_envvars_uppercase)
                config_value = read_envvar(self.envvar_prefix + config_name,
                                           force_uppercase=self.force_envvars_uppercase)
                if config_key_found is True:
                    # next the envvar if so desired to assist in matching to other file formats
                    if self.nest_envvars is True:
                        config_value = split_envvar(config_name, config_value, divider=self.envvar_divider)
                    break

            # will loop over all files of each type until all files of all types are searched, first time the key is
            # found will break outside of both loops
            elif config_type in self.valid_file_type_extension:
                for config_file in self.config_files_dict[config_type]:
                    if self.config_file_type == "file":
                        file_dict = self._parse_file_per_type(config_type, config_file)
                    else:
                        file_dict = self._parse_file_per_type(config_type, os.path.join(self.config_location,
                                                                                        config_file))
                    config_key_found, config_value = self._check_config_in_dict(config_name, file_dict)
                    if config_key_found is True:
                        break
                if config_key_found is True:
                    break
            else:
                raise ValueError

        # raise error if the key is required and not found in any of the config files, envvar or cli args
        if config_key_found is False and required is True:
            raise ValueError

        # if key is not required but still wasn't found take it from the key default value or failing that from the
        # global default value
        if config_key_found is False:
            if default_value is not None:
                config_value = default_value
            else:
                config_value = self.global_default_value

        # if type estimation is True try to guess the type of the value
        if self.type_estimate is True:
            config_value = estimate_type(config_value, none_values=self.none_values)

        # if the type the config is in the end isn't in the list of allowed_types and allowed_types list is set raise
        # a TypeError
        if allowed_types is not None:
            if type(config_value) not in allowed_types:
                raise TypeError

        return config_value

    def read_multiple_configuration_variables(self, config_names: list, default_value: Any = None,
                                              required: bool = False, allowed_types: Optional[list] = None) -> dict:
        """reads multiple keys of the configuration and returns the first value of each it found based on the priority
                of each config file option given in the __init__ of the class, basically a simple loop of the
                read_configuration_variable function with all the configurable values being the same in all iterations

                    Arguments:
                        config_names -- a list of the configuration key names you want to get the value of
                        default_value -- defaults to None, see config_type_priority in class init for it's use
                        required -- defaults to False, if set to True will ignore default_value & global_default_value
                            and will raise an ValueError if the configuration is not configured in any of the config
                            files/envvars/cli args
                        allowed_types -- Defaults to None, an optional list of types that are accepted for the variable
                            to be, if set a check will be preformed and if the variables value given is not of any of
                            the types in said list a TypeError will be raised
                    Returns:
                        config_value_dict -- a dict of the key/value pairs of all the configurations requested
        """

        config_value_dict = {}
        for config_name in config_names:
            config_value_dict[config_name] = self.read_configuration_variable(config_name, default_value, required,
                                                                              allowed_types)
        return config_value_dict

    def read_all_configuration_variables(self, default_value: Optional[dict] = None, required: Optional[list] = None,
                                         allowed_types: Optional[dict] = None) -> dict:
        """reads all configuration variables from all allowed sources and returns a dict that includes the combined
                        result of all of them, if a configuration variable exists in two (or more) different sources the
                         one with the higher priority will be the only one returned

                    Arguments:
                        default_value -- defaults to None, a dict of key/value pairs of a configuration variables & it's
                            value should it not be defined in any of the valid sources
                        required -- defaults to None, if given a list configuration variables it will raise a ValueError
                            if any of the configuration variables is not configured in any of the config
                            files/envvars/cli args
                        allowed_types -- Defaults to None, an optional dict of types that are accepted for a variable to
                            be, if set a check will be preformed and if the variables value given is not of any of the
                            types in said list a TypeError will be raised
                    Returns:
                        config_value_dict -- a dict of the key/value pairs of all the configurations requested
        """
        # first we create an empty config_value_dict
        config_value_dict = {}

        # now we fill the config_value_dict with the data of all valid sources in reverse order (from least desired to
        # the most desired source), overwriting each data that is found multiple times with the more desired state
        data_sources = self.config_type_priority
        data_sources.reverse()
        for config_type in data_sources:
            if config_type == "cli_args":
                config_value_dict.update(read_all_cli_args_to_dict())
            elif config_type == "envvars" or config_type == "env_vars":
                if self.nest_envvars is True:
                    config_value_dict.update(split_envvar_combained_dict(divider=self.envvar_divider,
                                                                         force_uppercase=self.force_envvars_uppercase))
                else:
                    config_value_dict.update(read_all_envvars_to_dict(force_uppercase=self.force_envvars_uppercase))
            # will loop over all files of each type until all files of all types are searched, first time the key is
            # found will break outside of both loops
            elif config_type in self.valid_file_type_extension:
                for config_file in self.config_files_dict[config_type]:
                    if self.config_file_type == "file":
                        file_dict = self._parse_file_per_type(config_type, config_file)
                    else:
                        file_dict = self._parse_file_per_type(config_type, os.path.join(self.config_location,
                                                                                        config_file))
                    config_value_dict.update(file_dict)
            else:
                raise ValueError

        # now we need to add the default values from the provided "default_value" dict to any configuration variable in
        # said list that wasn't found in any of the valid sources
        if default_value is not None:
            for default_config_key, default_config_value in default_value.items():
                config_found, config_value = self._check_config_in_dict(default_config_key, config_value_dict)
                if config_found is False:
                    config_value_dict[default_config_key] = default_config_value

        # and we run the type estimate (which is recursive) on the full dict if it's configured to be used
        if self.type_estimate is True:
            config_value_dict = estimate_type(config_value_dict, none_values=self.none_values)

        # now we check that all the required values exist and raise a ValueError otherwise
        if required is not None:
            for required_config in required:
                config_found, config_value = self._check_config_in_dict(required_config, config_value_dict)
                if config_found is False:
                    raise ValueError

        # and we also check that the "allowed_types" of all keys in the dict are from the list of allowed types and
        # raise a TypeError
        if allowed_types is not None:
            for allowed_types_key, allowed_types_value in allowed_types.items():
                if type(config_value_dict[allowed_types_key]) not in allowed_types_value:
                    raise TypeError

        # all that's left is returning the combined dict
        return config_value_dict

    @staticmethod
    def _check_config_in_dict(config_key: str, config_dict: dict) -> Tuple[bool, Any]:
        """internal function which checks if the key is in a given dict

            Arguments:
                config_key -- the configuration key name you want to check if it exists in the dict
                config_dict -- the dict you want to check if the is included in
            Returns:
                config_found -- True if the key is in the dict, false otherwise
                config_value -- the value of the configuration requested, returns None if the key is not part of the
                    the dict
        """

        if config_key in config_dict:
            config_value = config_dict[config_key]
            config_found = True
        else:
            config_value = None
            config_found = False
        return config_found, config_value

    def _parse_file_per_type(self, config_file_type: str, config_file_location: str) -> dict:
        """internal function which parses a file to a dict when given the file format type and it's location

            Arguments:
                config_file_type -- the type of the config file
                config_file_location -- the location of the config file
            Returns:
                file_dict -- a parsed dict of the config file data
        """

        if config_file_type in self.suffix_file_type_mapping["json"]:
            file_dict = parse_json_file(config_file_location)
        elif config_file_type in self.suffix_file_type_mapping["yaml"]:
            file_dict = parse_yaml_file(config_file_location)
        elif config_file_type in self.suffix_file_type_mapping["toml"]:
            file_dict = parse_toml_file(config_file_location)
        elif config_file_type in self.suffix_file_type_mapping["ini"]:
            file_dict = parse_ini_file(config_file_location)
        elif config_file_type in self.suffix_file_type_mapping["hcl"]:
            file_dict = parse_hcl_file(config_file_location)
        elif config_file_type in self.suffix_file_type_mapping["xml"]:
            file_dict = parse_xml_file(config_file_location)
        elif config_file_type in self.suffix_file_type_mapping["env"]:
            file_dict = parse_env_file(config_file_location)
        else:
            raise ValueError
        return file_dict



import os
import dpath.util
import dpath.options
from typing import Optional, Union


def read_envvar(envvar: str, force_uppercase: bool = True) -> Optional[str]:
    """Read an environment variable, if force_uppercase is true will convert all keys
        to be UPPERCASE

            Arguments:
                envvar -- name of the envvar to get the value of
                force_uppercase -- if the envvar key will be forced to be all in UPPERCASE, defaults to True
            Returns:
                the value of the envvar, None if doesn't exist
    """
    if force_uppercase is True:
        envvar = envvar.upper()
    envvar_value = os.getenv(envvar)

    if isinstance(envvar_value, str) is True:
        # this weird encode and decode is to avoid some cases where envvar get special characters escaped
        envvar_value = envvar_value.encode('latin1').decode('unicode_escape')
    return envvar_value


def envvar_defined(envvar: str, force_uppercase: bool = True) -> bool:
    """Check if an environment variable is defined, if force_uppercase is true will convert all keys
        to be UPPERCASE

            Arguments:
                envvar -- name of the envvar to get the value of
                force_uppercase -- if the envvar key will be forced to be all in UPPERCASE, defaults to True
            Returns:
                True if envvar is declared, False otherwise
    """
    if force_uppercase is True:
        envvar = envvar.upper()

    if envvar in os.environ:
        return True
    else:
        return False


def read_all_envvars_to_dict(force_uppercase: bool = True) -> dict:
    """Read all environment variables and return them in a dict form, if force_uppercase is true will convert all keys
        to be UPPERCASE

            Arguments:
                force_uppercase -- while counter-intuitive in the naming it means that if the environment variable
                    is uppercase the dict will treat it as the same one as a lowercase one & will return it in
                    lowercase form (name saved to match all the other uses of said function)
            Returns:
                envvar_dict -- A dict of all environment variables key/value pairs
    """
    envvar_dict = {}
    for envvar in os.environ:
        if force_uppercase is True:
            envvar_dict[envvar.lower()] = os.environ.get(envvar)
        else:
            envvar_dict[envvar] = os.environ.get(envvar)
    return envvar_dict


def split_envvar(envvar: Union[str, list], value: str, divider: str = "_"):
    """Take an envvar & it's value and split it by the divider to a nested dictionary

                Arguments:
                    envvar -- the envvar key to split into nested dictionary
                    value -- the bottom most value of the nested envvars keys
                    divider -- the string letter by which to divide the envvar key by, defaults to "_"
                Returns:
                    envvar_dict -- A dict that is the result of the envvar being split by the divider with the value
                        appended as the bottom most of the nest key
    """
    if type(envvar) == str:
        envvar_list = envvar.split(divider)
    else:
        envvar_list = envvar

    if len(envvar_list) > 1:
        envvar_dict = {
            envvar_list[0]: split_envvar(envvar_list[1:], value, divider)
        }
    else:
        envvar_dict = {
            envvar_list[0]: value
        }

    return envvar_dict


def split_envvar_combained_dict(divider: str = "_", force_uppercase: bool = True):
    """Returns a dict of all envvars that has had their keys split by the divider into nested dicts

                    Arguments:
                        divider -- the string letter by which to divide the envvar key by, defaults to "_"
                        force_uppercase -- if the envvar key will be forced to be all in UPPERCASE, defaults to True
                    Returns:
                        envvar_split_dict -- A dict that is the result of all envvars being split by the divider with
                            the value appended as the bottom most of the nest key
    """
    envvar_dict = read_all_envvars_to_dict(force_uppercase=force_uppercase)
    envvar_split_dict = {}
    for envvar_key, envvar_value in envvar_dict.items():
        dpath.options.ALLOW_EMPTY_STRING_KEYS=True
        temp_split_envvar = split_envvar(envvar_key, envvar_value, divider=divider)
        dpath.util.merge(envvar_split_dict, temp_split_envvar)
    return envvar_split_dict



from configobj import ConfigObj


def parse_ini_file(path_to_ini_file: str) -> dict:
    """take a path to a INI file & returns it as a valid python dict.

            Arguments:
                path_to_ini_file -- the path of the ini file
            Returns:
                config_file_dict -- dict of the file
    """
    config = ConfigObj(path_to_ini_file)
    return dict(config)



from dotenv import dotenv_values


def parse_env_file(path_to_env_file: str) -> dict:
    """take a path to a env file & returns it as a valid python dict.

            Arguments:
                path_to_env_file -- the path of the json file
            Returns:
                config_file_dict -- dict of the file
    """
    return dotenv_values(dotenv_path=path_to_env_file)



from parse_it.file.file_reader import *
import hcl


def parse_hcl_file(path_to_hcl_file: str) -> dict:
    """take a path to a HCL file & returns it as a valid python dict.

            Arguments:
                path_to_hcl_file -- the path of the hcl file
            Returns:
                config_file_dict -- dict of the file
    """
    return hcl.loads(read_file(path_to_hcl_file))



from parse_it.file.file_reader import *
import xmltodict
import json


def parse_xml_file(path_to_xml_file: str) -> dict:
    """take a path to a XML file & returns it as a valid python dict.

            Arguments:
                path_to_xml_file -- the path of the xml file
            Returns:
                config_file_dict -- dict of the file
    """
    # the dump & load to/from JSON is to change it from ordered dict to a normal dict
    return json.loads(json.dumps(xmltodict.parse(read_file(path_to_xml_file))))



from setuptools import setup, find_packages
from datetime import datetime


__author__ = 'Naor Livne'
__author_email__ = 'naorlivne@gmail.com'
__version__ = datetime.now().strftime("%Y.%m.%d.%H.%M")

# read the README.md file for the long description of the package
with open('README.md') as f:
    long_description = f.read()

# minimum requirement list
requirements = [
    "PyYAML",
    "toml",
    "configobj",
    "xmltodict",
    "pyhcl",
    "python-dotenv",
    "dpath"
]

# optional requirements, typing is used for support of Python versions 3.4 & lower, note 3.4 and lower is untested
extra_requirements = {
          'typing':  ["typing"]
      }

setup(
    name='parse_it',
    author=__author__,
    author_email=__author_email__,
    version=__version__,
    description="A python library for parsing multiple types of config files, envvars and command line arguments "
                "which takes the headache out of setting app configurations.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    extras_require=extra_requirements,
    scripts=['setup.py'],
    license="LGPLv3",
    keywords="parse_it parsing parse parser yaml json xml toml ini cfg hcl envvar environment_variable config "
             "cli_args command_line_arguments tml yml configuration configuration_file",
    url="https://github.com/naorlivne/parse_it",
    install_requires=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Other Environment",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
        ]
)
